/**
    @file
   @author Shin'ichiro Nakaoka
*/

#include "SceneItem.h"
#include "ItemTreeView.h"
#include "ItemManager.h"
#include "Archive.h"
#include "RootItem.h"
#include "OptionManager.h"
#include "PutPropertyFunction.h"
#include <cnoid/YAMLReader>
#include <cnoid/FileUtil>
#include <boost/bind.hpp>
#include "gettext.h"

using namespace std;
using namespace boost;
using namespace cnoid;

namespace {

    const bool TRACE_FUNCTIONS = false;

    bool loadSceneItem(SceneItem* item, const std::string& filename, std::ostream& os)
    {
        SceneObject::ReadResult result = item->loadScene(filename);
        if(result.success()){
            if(item->name().empty()){
                item->setName(getFilename(filesystem::path(filename)));
            }
        } else {
            os << result.message();
        }

        return result.success();
    }


    class SceneItemManager
    {
        std::set<SceneItemPtr> visibleSceneItems;

    public:

        SceneItemManager(ExtensionManager* ext) {

            ext->itemManager().registerClass<SceneItem>("SceneItem");

            ext->itemManager().addLoader<SceneItem>(
                _("Scene file"), "OSG-IMPORTABLE-SCENE-FILE", "", bind(loadSceneItem, _1, _2, _3), ItemManager::PRIORITY_CONVERSION);

            ext->optionManager().addOption("model", program_options::value< vector<string> >(), "load a 3d model file");
            ext->optionManager().sigOptionsParsed().connect(
                bind(&SceneItemManager::onSigOptionsParsed, this, _1));

            ItemTreeView::mainInstance()->sigCheckToggled().connect(
                bind(&SceneItemManager::onItemCheckToggled, this, _1, _2));
        }

        void onSigOptionsParsed(boost::program_options::variables_map& variables) {

            if(variables.count("model")){
                vector<string> files = variables["model"].as< vector<string> >();
                for(size_t i=0; i < files.size(); ++i){
                    SceneItemPtr item(new SceneItem());
                    if(item->load(files[i], "OSG-IMPORTABLE-SCENE-FILE")){
                        RootItem::mainInstance()->addChildItem(item);
                        ItemTreeView::mainInstance()->checkItem(item, true);
                    }
                }
            }
        }

        void onItemCheckToggled(Item* item, bool isChecked) {
            SceneItem* sceneItem = dynamic_cast<SceneItem*>(item);
            if(sceneItem){
                std::set<SceneItemPtr>::iterator p = visibleSceneItems.find(sceneItem);
                if(isChecked){
                    if(p == visibleSceneItems.end()){
                        sceneItem->sceneObject()->show();
                        visibleSceneItems.insert(sceneItem);
                    }
                } else {
                    if(p != visibleSceneItems.end()){
                        sceneItem->sceneObject()->hide();
                        visibleSceneItems.erase(p);
                    }
                }
            }
        }

        ~SceneItemManager() {
            std::set<SceneItemPtr>::iterator p;
            for(p = visibleSceneItems.begin(); p != visibleSceneItems.end(); ++p){
                SceneItemPtr sceneItem = *p;
                sceneItem->sceneObject()->hide();
            }
        }
    };
}
    

void SceneItem::initialize(ExtensionManager* ext)
{
    static bool initialized = false;
    if(!initialized){
        ext->manage(new SceneItemManager(ext));
        initialized = true;
    }
}


SceneItem::SceneItem()
    : sceneObject_(new SceneObject)
{
    setAttribute(LOAD_ONLY);
}


SceneItem::SceneItem(const SceneItem& org)
    : Item(org),
      sceneObject_(new SceneObject(*org.sceneObject_))
{

}


SceneItem::~SceneItem()
{

}


void SceneItem::onDisconnectedFromRoot()
{
    sceneObject_->hide();
}


SceneObject::ReadResult SceneItem::loadScene(const std::string& filename)
{
    return sceneObject_->load(filename);
}


ItemPtr SceneItem::doDuplicate() const
{
    return new SceneItem(*this);
}


void SceneItem::doPutProperties(PutPropertyFunction& putProperty)
{
    putProperty(_("Scene file"), getFilename(filesystem::path(lastAccessedFilePath())));
}


bool SceneItem::store(Archive& archive)
{
    if(!lastAccessedFilePath().empty()){
        archive.writeRelocatablePath("file", lastAccessedFilePath());
        archive.write("format", lastAccessedFileFormatId());
    }
    return true;
}


bool SceneItem::restore(const Archive& archive)
{
    std::string filename, formatId;
    if(archive.readRelocatablePath("file", filename) && archive.read("format", formatId)){
        if(load(filename, archive.currentParentItem(), formatId)){
            return true;
        }
    }
    return false;
}
