/**
   @file
   @author Shin'ichiro NAKAOKA
*/

#include "BodyMotion.h"
#include "Body.h"
#include "Link.h"
#include <cnoid/YAMLReader>
#include <cnoid/YAMLWriter>

using namespace std;
using namespace boost;
using namespace cnoid;

namespace {
    //bool TRACE_FUNCTIONS = false;
}


BodyMotion::BodyMotion()
    : AbstractMultiSeq("BodyMotion"),
      jointPosSeq_(new MultiValueSeq()),
      linkPosSeq_(new MultiSE3Seq())
{
    jointPosSeq_->setSeqContentName("JointPosition");
    linkPosSeq_->setSeqContentName("LinkPosition");
}


BodyMotion::BodyMotion(const BodyMotion& org)
    : AbstractMultiSeq(org),
      jointPosSeq_(new MultiValueSeq(*org.jointPosSeq_)),
      linkPosSeq_(new MultiSE3Seq(*org.linkPosSeq_))
{
    if(org.ZMPseq_){
        ZMPseq_.reset(new Vector3Seq(*org.ZMPseq_));
    }
}


const Vector3SeqPtr& BodyMotion::ZMPseq()
{
    if(!ZMPseq_){
        ZMPseq_.reset(new Vector3Seq());
        ZMPseq_->setSeqContentName("ZMP");
        ZMPseq_->setNumFrames(numFrames());
        ZMPseq_->setFrameRate(frameRate());
    }
    return ZMPseq_;
}


/*
  This function sets the number of joints
*/
void BodyMotion::setNumParts(int numParts, bool clearNewElements)
{
    jointPosSeq_->setNumParts(numParts, clearNewElements);
}


/**
   This function returns the number of joints
*/
int BodyMotion::getNumParts() const
{
    return jointPosSeq_->numParts();
}
        

double BodyMotion::getFrameRate() const
{
    return frameRate();
}


void BodyMotion::setFrameRate(double frameRate)
{
    jointPosSeq_->setFrameRate(frameRate);
    linkPosSeq_->setFrameRate(frameRate);
    if(ZMPseq_){
        ZMPseq_->setFrameRate(frameRate);
    }
}


int BodyMotion::getNumFrames() const
{
    return numFrames();
}


void BodyMotion::setNumFrames(int n, bool clearNewArea)
{
    jointPosSeq_->setNumFrames(n, clearNewArea);
    linkPosSeq_->setNumFrames(n, clearNewArea);
    if(ZMPseq_){
        ZMPseq_->setNumFrames(n, clearNewArea);
    }
}


void BodyMotion::setDimension(int numFrames, int numJoints, int numLinks, bool clearNewArea)
{
    jointPosSeq_->setDimension(numFrames, numJoints, clearNewArea);
    linkPosSeq_->setDimension(numFrames, numLinks, clearNewArea);
    if(ZMPseq_){
        ZMPseq_->setNumFrames(numFrames, clearNewArea);
    }
}


void BodyMotion::setDimension(int numFrames, int numJoints, bool clearNewArea)
{
    jointPosSeq_->setDimension(numFrames, numJoints, clearNewArea);
    linkPosSeq_->setNumFrames(numFrames, clearNewArea);
    if(ZMPseq_){
        ZMPseq_->setNumFrames(numFrames, clearNewArea);
    }
}    


bool BodyMotion::loadStandardYAMLformat(const std::string& filename)
{
    bool result = false;
    clearSeqMessage();
    YAMLReader reader;
    reader.expectRegularMultiListing();
    
    try {
        reader.load(filename);
        const Mapping& archive = *reader.document()->toMapping();
        result = read(archive);

    } catch(const ValueNode::Exception& ex){
        result = false;
        addSeqMessage(ex.message());
    }

    return result;
}


bool BodyMotion::read(const Mapping& archive)
{
    setDimension(0, 1, 1);

    bool result = true;
    bool loaded = false;
    bool zmpLoaded = false;

    try {
        if(archive["type"].toString() != "BodyMotion"){
            result = false;

        } else {
            const Listing& components = *archive["components"].toListing();
        
            for(int i=0; i < components.size(); ++i){
                const Mapping& component = *components[i].toMapping();
                const string& type = component["type"];
                string content;
                if(!component.read("content", content)){
                    component.read("purpose", content);
                }
                if(type == "MultiValueSeq" && content == "JointPosition"){
                    result &= jointPosSeq_->readSeq(component);
                    if(result){
                        loaded = true;
                    } else {
                        addSeqMessage(jointPosSeq_->seqMessage());
                    }
                } else if((type == "MultiAffine3Seq" || type == "MultiSe3Seq" || type == "MultiSE3Seq")
                          && content == "LinkPosition"){
                    result &= linkPosSeq_->readSeq(component);
                    if(result){
                        loaded = true;
                    } else {
                        addSeqMessage(linkPosSeq_->seqMessage());
                    }
                } else if(type == "Vector3Seq" && content == "ZMP"){
                    result = ZMPseq()->readSeq(component);
                    if(result){
                        zmpLoaded = true;
                    } else {
                        addSeqMessage(ZMPseq()->seqMessage());
                    }
                }
                if(!result){
                    break;
                }
            }
        }
    } catch(const ValueNode::Exception& ex){
        addSeqMessage(ex.message());
        result = false;
    }

    if(!result){
        setDimension(0, 1, 1);
    }
    if(!zmpLoaded){
        ZMPseq_.reset();
    }
    
    return (result && loaded);
}


bool BodyMotion::write(YAMLWriter& writer)
{
    writer.startListing();

    if(jointPosSeq_->numFrames() > 0){
        if(!jointPosSeq_->writeSeq(writer)){
            addSeqMessage(jointPosSeq_->seqMessage());
            return false;
        }
    }
    if(linkPosSeq_->numFrames() > 0){
        if(!linkPosSeq_->writeSeq(writer)){
            addSeqMessage(linkPosSeq_->seqMessage());
            return false;
        }
    }
    if(hasZMPseq()){
        if(!ZMPseq_->writeSeq(writer)){
            addSeqMessage(ZMPseq_->seqMessage());
            return false;
        }
    }

    writer.endListing();

    return true;
}


bool BodyMotion::saveAsStandardYAMLformat(const std::string& filename)
{
    YAMLWriter writer(filename);
    writer.setDoubleFormat("%.9g");

    writer.putComment("Body motion data set format version 1.0 defined by cnoid-Robotics\n");

    writer.startMapping();

    writer.putKeyValue("type", "BodyMotion");
    writer.putKey("components");

    bool result = write(writer);

    writer.endMapping();

    return result;
}


namespace cnoid {

    BodyMotion::Frame& operator<<(const BodyMotion::Frame& frame, const Body& body)
    {
        BodyMotion& motion = const_cast<BodyMotion::Frame&>(frame).motion();
        int numJoints = std::min(body.numJoints(), motion.numJoints());
        MultiValueSeq::Frame q = motion.jointPosSeq()->frame(frame.frame());
        for(int i=0; i < numJoints; ++i){
            q[i] = body.joint(i)->q;
        }
        int numLinks =  std::min(body.numLinks(), motion.numLinks());
        MultiSE3Seq::Frame p = motion.linkPosSeq()->frame(frame.frame());
        for(int i=0; i < numLinks; ++i){
            Link* link = body.link(i);
            p[i].set(link->p, link->R);
        }
        return const_cast<BodyMotion::Frame&>(frame);
    }

    BodyMotion::Frame& operator>>(const BodyMotion::Frame& frame, const Body& body)
    {
        const BodyMotion& motion = frame.motion();
        int numJoints = std::min(body.numJoints(), motion.numJoints());
        const MultiValueSeq::Frame q = motion.jointPosSeq()->frame(frame.frame());
        for(int i=0; i < numJoints; ++i){
            body.joint(i)->q = q[i];
        }
        int numLinks =  std::min(body.numLinks(), motion.numLinks());
        const MultiSE3Seq::Frame p = motion.linkPosSeq()->frame(frame.frame());
        for(int i=0; i < numLinks; ++i){
            Link* link = body.link(i);
            link->p = p[i].translation();
            link->R = p[i].rotation();
        }
        return const_cast<BodyMotion::Frame&>(frame);
    }
}
